!
! IonogramDerivedData.f90 -- data structures holding ionogram-derived data.
!
! Version: v 0.2 2010/11/02
! 
! Ivan Galkin
! University of Massachusetts Lowell
! Center for Atmospheric Research
! Email: Ivan_Galkin@uml.edu
!
!  Edited 2010/11/02 by Mark A. Hausman: added (nearly) complete initialization
!     of idd fields in subroutine idd_reset.
!
! Sample design for use with the SAO.XML read/write library.
! Data from SAO.XML input files is put into these data structures
! Data for SAO.XML output files is taken from these data structures
! This is only an example. 
! Actual design of data architecture is expected to be application-specific.
!
! Use of real(8) was necessary to avoid precision effects in the SAOXML formatted output
!
!
module IonogramDerivedData
   implicit none

   !
   ! ======================   c o n s t a n t s      ==================================
   !
   !
   ! by convention, constants are written in all caps with underscores separating words
   !
   integer, parameter               :: IDD_MAX_STATION_NAME     = 100
   integer, parameter               :: IDD_MAX_COMMENTS_LENGTH  = 255
   integer, parameter               :: IDD_MAX_NAME             = 100
   integer, parameter               :: IDD_MAX_VERSION          = 100
   integer, parameter               :: IDD_MAX_DESCRIPTION      = 255
   integer, parameter               :: IDD_MAX_UNITS_NAME       = 20
   integer, parameter               :: IDD_MAX_CHAR_NAME        = 20
   integer, parameter               :: IDD_MAX_TYPE_NAME        = 20
   integer, parameter               :: IDD_MAX_LAYER_NAME       = 10
   integer, parameter               :: IDD_MAX_MODEL_NAME       = 20
   integer, parameter               :: IDD_MAX_VALUE_NAME       = 100

   real(8), private                 :: a
   real(8), parameter               :: NO_VALUE_CHAR = huge(a)
   integer, parameter               :: NO_VALUE_INDEX = -1
   integer, parameter               :: NO_VALUE_YEAR = -99999
   


   ! List of enumerated standard URSI characteristics
   ! Enumerated from 1 to TOTAL_STANDARD_CHARS
   ! See list of standard URSI characteristics in SAO.XML description, Appendix C.

   integer, parameter               :: I_FOF2      = 1
   integer, parameter               :: I_FXF2      = 2
   integer, parameter               :: I_FZF2      = 3
   integer, parameter               :: I_M3000F2   = 4
   integer, parameter               :: I_HMINF2    = 5
   integer, parameter               :: I_HP2       = 6
   integer, parameter               :: I_HMINOX    = 7
   integer, parameter               :: I_MUF3000F2 = 8
   integer, parameter               :: I_HC        = 9
   integer, parameter               :: I_QC        = 10
   integer, parameter               :: I_FOF1      = 11
   integer, parameter               :: I_FXF1      = 12
   integer, parameter               :: I_M3000F1   = 13
   integer, parameter               :: I_HMINF1    = 14
   integer, parameter               :: I_HMINF     = 15
   integer, parameter               :: I_MUF3000F1 = 16
   integer, parameter               :: I_FOE       = 17
   integer, parameter               :: I_FOE2      = 18
   integer, parameter               :: I_FOEA      = 19
   integer, parameter               :: I_HMINE     = 20
   integer, parameter               :: I_HMINE2    = 21
   integer, parameter               :: I_HMINEA    = 22
   integer, parameter               :: I_FOES      = 23
   integer, parameter               :: I_FXES      = 24
   integer, parameter               :: I_FBES      = 25
   integer, parameter               :: I_FTES      = 26
   integer, parameter               :: I_HMINES    = 27
   ! Type-Es is not included here, it is not a floating point value, but rather a text string
   integer, parameter               :: I_FOF1P5    = 28
   integer, parameter               :: I_FMIN      = 29
   integer, parameter               :: I_M3000F1P5 = 30
   integer, parameter               :: I_HMINF1P5  = 31
   integer, parameter               :: I_FM2       = 32
   integer, parameter               :: I_HM        = 33
   integer, parameter               :: I_FM3       = 34
   integer, parameter               :: I_FOI       = 35
   integer, parameter               :: I_FXI       = 36
   integer, parameter               :: I_FMI       = 37
   integer, parameter               :: I_M3000I    = 38
   integer, parameter               :: I_HMINI     = 39
   integer, parameter               :: I_FOP       = 40
   integer, parameter               :: I_HMINP     = 41
   integer, parameter               :: I_DFS       = 42
   integer, parameter               :: I_FHMINF2   = 43
   integer, parameter               :: I_FHMINF    = 44
   integer, parameter               :: I_HMINMF1   = 45
   integer, parameter               :: I_H1        = 46
   integer, parameter               :: I_H2        = 47
   integer, parameter               :: I_H3        = 48
   integer, parameter               :: I_H4        = 49
   integer, parameter               :: I_H5        = 50
   integer, parameter               :: I_H         = 51
   integer, parameter               :: I_I2000     = 52
   integer, parameter               :: I_I         = 53
   integer, parameter               :: I_I1000     = 54
   integer, parameter               :: I_T         = 55
   integer, parameter               :: I_FMINF     = 56
   integer, parameter               :: I_FMINE     = 57
   integer, parameter               :: I_HOM       = 58
   integer, parameter               :: I_YE        = 59
   integer, parameter               :: I_QF        = 60
   integer, parameter               :: I_QE        = 61
   integer, parameter               :: I_FF        = 62
   integer, parameter               :: I_FE        = 63
   integer, parameter               :: I_FMUF3000  = 64
   integer, parameter               :: I_HMINMUF3000 = 65
   integer, parameter               :: I_ZME       = 66
   integer, parameter               :: I_ZMF1      = 67
   integer, parameter               :: I_ZMF2      = 68
   integer, parameter               :: I_ZHALFNM   = 69
   integer, parameter               :: I_YF2       = 70
   integer, parameter               :: I_YF1       = 71
   integer, parameter               :: I_B0        = 72
   integer, parameter               :: I_B1        = 73
   integer, parameter               :: I_D1        = 74

   integer, parameter               :: IDD_TOTAL_STANDARD_CHARS = 74

   ! the following is used to convert standard URSI 2-letter codes of characteristics
   ! to the appropriate index to the array of characteristics "char"
   ! make sure URSI codes match indeces above

   character*2, parameter           :: URSI_CODE_CHAR(IDD_TOTAL_STANDARD_CHARS) = (/ &
                                       '00', '01', '02', '03', '04', '05', '06', '07', '08', '09', & 
                                       '10', '11', '13', '14', '16', '17', & 
                                       '20', '22', '23', '24', '26', '27', & 
                                       '30', '31', '32', '33', '34', & 
                                       '40', '42', '43', '44', '47', '48', '49', & 
                                       '50', '51', '52', '53', '54', '55', '56', '57', & 
                                       '60', '61', '63', '64', '65', '66', '67', '68', '69', & 
                                       '70', '71', '72', '79', & 
                                       '80', '81', '82', '83', '84', '85', '86', '87', '88', '89', & 
                                       '90', '91', '92', '93', '94', '95', & 
                                       'D0', 'D1', 'D2' &
                                       /)

   character(len=*), parameter      :: URSI_NAMES_CHAR(IDD_TOTAL_STANDARD_CHARS) = (/ &
                                       'foF2     ','fxF2     ','fzF2     ','M3000F2  ',"h'F2     ",'hpF2     ', &
                                       "h'Ox     ","MUF3000F2",'hc       ','qc       ', &
                                       'foF1     ','fxF1     ','M3000F1  ',"h'F1     ","h'F      ",'MUF3000F1', &
                                       'foE      ','foE2     ','foEa     ',"h'E      ","h'E2     ","h'Ea     ", &
                                       'foEs     ','fxEs     ','fbEs     ','ftEs     ',"h'Es     ", &
                                       'foF1.5   ','fmin     ','M3000F1.5',"h'F1.5   ",'fm2      ','hm       ','fm3      ', &
                                       'foI      ','fxI      ','fmI      ','M3000I   ',"h'I      ", & 
                                       'foP      ',"h'P      ",'dfs      ', &
                                       "fh'F2    ","fh'F     ","h'mF1    ",'h1       ','h2       ', &
                                       'h3       ','h4       ','h5       ','H        ', &
                                       'I2000    ','I        ','I1000    ','T        ', &
                                       'fminF    ','fminE    ','HOM      ','yE       ','QF       ', &
                                       'QE       ','FF       ','FE       ','fMUF3000 ',"h'MUF3000", &
                                       'zmE      ','zmF1     ','zmF2     ','zhalfNm  ','yF2      ','yF1      ', &
                                       'B0       ','B1       ','D1       ' &
                                       /)

   character(len=*), parameter      :: URSI_UNITS_CHAR(IDD_TOTAL_STANDARD_CHARS) = (/ &
                                       'MHz ', 'MHz ', 'MHz ', '    ', 'km  ', 'km  ', 'km  ', 'MHz ', 'km  ', 'km  ', & 
                                       'MHz ', 'MHz ', '    ', 'km  ', 'km  ', 'MHz ', & 
                                       'MHz ', 'MHz ', 'MHz ', 'km  ', 'km  ', 'km  ', & 
                                       'MHz ', 'MHz ', 'MHz ', 'MHz ', 'km  ', & 
                                       'MHz ', 'MHz ', '    ', 'km  ', 'MHz ', 'km  ', 'MHz ', & 
                                       'MHz ', 'MHz ', 'MHz ', '    ', 'km  ', 'MHz ', 'km  ', 'MHz ', & 
                                       'MHz ', 'MHz ', 'km  ', 'km  ', 'km  ', 'km  ', 'km  ', 'km  ', 'km  ', & 
                                       'TECU', 'TECU', 'TECU', 'TECU', & 
                                       'MHz ', 'MHz ', 'km  ', 'km  ', 'km  ', 'km  ', 'MHz ', 'MHz ', 'MHz ', 'km  ', & 
                                       'km  ', 'km  ', 'km  ', 'km  ', 'km  ', 'km  ', & 
                                       'km  ', '    ', '    ' &
                                       /)

   character(len=*), parameter      :: NO_VALUE_UNITS = 'unknown'

   ! Model characteristics that are stored together with measured data for comparisons

   integer, parameter               :: I_FOF2P     = 1 + IDD_TOTAL_STANDARD_CHARS
   integer, parameter               :: I_FOF1P     = 2 + IDD_TOTAL_STANDARD_CHARS
   integer, parameter               :: I_FOEP      = 3 + IDD_TOTAL_STANDARD_CHARS

   integer, parameter               :: IDD_TOTAL_MODELED_CHARS = 3

   character (len=*), parameter     :: MODELED_CHAR_NAME(IDD_TOTAL_MODELED_CHARS) = (/ &
                                      'foF2p', & 
                                      'foF1p', & 
                                      'foEp '  &
                                      /)

   ! Custom chars for Digisonde (example)
   ! Things like foF3 should go here
   ! Newly added chars shall be first checked for conflicts with existing custom char names

   integer, parameter               :: I_DOWNF     = 1 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
   integer, parameter               :: I_DOWNE     = 2 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
   integer, parameter               :: I_DOWNES    = 3 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
   integer, parameter               :: I_D         = 4 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
   integer, parameter               :: I_DLT_FOF2  = 5 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
   integer, parameter               :: I_FMINES    = 6 + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS

   integer, parameter               :: IDD_TOTAL_CUSTOM_CHARS = 6

   character (len=*), parameter     :: CUSTOM_CHAR_NAME(IDD_TOTAL_CUSTOM_CHARS) = (/ &
                                       'DownF   ', &
                                       'DownE   ', &
                                       'DownEs  ', &
                                       'D       ', &
                                       'dlt foF2', &
                                       'fminEs  ' &
                                       /)

   ! Total number of characteristics stored

   integer, parameter               :: IDD_TOTAL_CHARS = &
                                         IDD_TOTAL_STANDARD_CHARS + &
                                         IDD_TOTAL_MODELED_CHARS + &
                                         IDD_TOTAL_CUSTOM_CHARS

   ! storage of traces
   integer, parameter               :: IDD_MAX_TRACES             = 20

   ! storage of profiles
   integer, parameter               :: IDD_MAX_PROFILES           = 20

   ! system constants
   integer, parameter               :: IDD_UNKNOWN_STEPPING       = 0
   integer, parameter               :: IDD_LINEAR_STEPPING        = 1
   integer, parameter               :: IDD_LOG_STEPPING           = 2
   integer, parameter               :: IDD_TABULATED_STEPPING     = 3


   !
   ! ======================   d a t a    t y p e s      ==================================
   !
   !
   ! Define data types to store ionogram-derived data elements
   !
   !
   type IDD_TIMESTAMP_UTC
      integer          :: year               ! 4 digits
      integer          :: month              ! 1..12
      integer          :: day_of_month       ! 1..31
      integer          :: day_of_year        ! 1..365
      integer          :: hour               ! 0..23
      integer          :: minute             ! 0..59
      integer          :: second             ! 0..59
      integer          :: millisecond        ! 0..999
   end type IDD_TIMESTAMP_UTC
    
   type IDD_LOCATION
      character(len=IDD_MAX_STATION_NAME)    :: station_name        
      character*5                            :: ursi_code
      real(8)                                :: geo_latitude        
      real(8)                                :: geo_longitude
      integer                                :: uml_sid
   end type IDD_LOCATION

   type IDD_IONOGRAM_MODE
      real(8)                                :: start_frequency_MHz
      real(8)                                :: stop_frequency_MHz
      integer                                :: frequency_stepping_rule
      real(8)                                :: frequency_step
      character(len=IDD_MAX_UNITS_NAME)      :: frequency_step_units
      ! ToDo: tabulated frequency stepping
      real(8)                                :: start_range_km
      real(8)                                :: stop_range_km
      integer                                :: range_stepping_rule
      real(8)                                :: range_step
      character(len=IDD_MAX_UNITS_NAME)      :: range_step_units
      ! restricted frequencies
      integer                                :: total_restricted_freq_bands
      real(8), dimension(:), pointer         :: lower_restricted_freq_limits
      real(8), dimension(:), pointer         :: upper_restricted_freq_limits
      character(len=IDD_MAX_UNITS_NAME)      :: restricted_freq_units
   end type IDD_IONOGRAM_MODE
   
   type IDD_CONTACT_PERSON
      logical                                :: available
      character(len=IDD_MAX_NAME)            :: name
      character(len=IDD_MAX_DESCRIPTION)     :: affiliation
      character(len=IDD_MAX_DESCRIPTION)     :: address
      character(len=IDD_MAX_DESCRIPTION)     :: email
   end type IDD_CONTACT_PERSON

   type IDD_SYSTEM
      logical                                :: autoscaled
      character(len=IDD_MAX_NAME)            :: manual_scaler
      character(len=IDD_MAX_NAME)            :: auto_scaler
      character(len=IDD_MAX_VERSION)         :: auto_scaler_version
      integer, dimension(20)                 :: artist_flags
      character(len=IDD_MAX_COMMENTS_LENGTH) :: comments
      character(len=IDD_MAX_NAME)            :: source
      character(len=IDD_MAX_NAME)            :: source_type ! i.e., ionosonde model
      type(IDD_IONOGRAM_MODE)                :: ionogram
      type(IDD_CONTACT_PERSON)               :: contact
      character(len=IDD_MAX_NAME)            :: digisonde_preface ! custom-added
      character(len=IDD_MAX_NAME),dimension(IDD_TOTAL_MODELED_CHARS) :: model_name
      character(len=IDD_MAX_NAME),dimension(IDD_TOTAL_MODELED_CHARS) :: model_options
   end type IDD_SYSTEM

   type IDD_SOLAR_TERRESTRIAL_DATA
      real(8)                                :: sunspot_number
      character(len=IDD_MAX_NAME)            :: sunspot_number_status
      real(8)                                :: gyro_frequency_MHz
      real(8)                                :: dip_angle_deg
      character(len=IDD_MAX_NAME)            :: magfield_model_name
      real(8)                                :: magfield_model_altitude
   end type IDD_SOLAR_TERRESTRIAL_DATA

   type IDD_CHAR
      real(8)                                :: value
      real(8)                                :: bound
      real(8)                                :: lower_bound
      real(8)                                :: upper_bound
      character(len=IDD_MAX_UNITS_NAME)      :: boundary_type
      character                              :: q
      character                              :: d
      character(len=IDD_MAX_UNITS_NAME)      :: units
      character(len=IDD_MAX_DESCRIPTION)     :: description
      character                              :: flag ! edited or validated
   end type IDD_CHAR

   type IDD_TRACE
      logical                                :: available
      character(len=IDD_MAX_TYPE_NAME)       :: trace_type
      character(len=IDD_MAX_LAYER_NAME)      :: layer
      integer                                :: multiple
      character                              :: polarization
      integer                                :: num
      real(8), dimension(:), pointer         :: frequencies
      real(8), dimension(:), pointer         :: ranges
      logical                                :: amplitudes_available
      real(8)                                :: amplitude_no_value
      real(8), dimension(:), pointer         :: amplitudes
      logical                                :: doppler_shifts_available
      real(8)                                :: doppler_shift_no_value
      real(8), dimension(:), pointer         :: doppler_shifts
     ! other trace values can be added here
   end type IDD_TRACE

   type IDD_PROFILE_COEFS
      character(len=IDD_MAX_LAYER_NAME)      :: region
      real(8)                                :: start_freq_MHz
      real(8)                                :: end_freq_MHz
      real(8)                                :: peak_height_km
      real(8)                                :: z_half_Nm_km
      integer                                :: num
      real(8), dimension(:), pointer         :: c
      real(8)                                :: fit_error
   end type IDD_PROFILE_COEFS

   type IDD_QP_SEGMENT
      integer                                :: id
      real(8)                                :: start_distance_km
      real(8)                                :: end_distance_km
      real(8)                                :: a
      real(8)                                :: b
      real(8)                                :: c
      real(8)                                :: fit_error
   end type IDD_QP_SEGMENT

   type IDD_TOPSIDE_CHAPMAN_LAYER
      logical                                :: available
      character(len=IDD_MAX_DESCRIPTION)     :: model
      real(8)                                :: peak_height_km
      real(8)                                :: peak_density_cm3
      real(8)                                :: peak_scale_height_km
   end type IDD_TOPSIDE_CHAPMAN_LAYER

   type IDD_PROFILE
      
      logical                                :: available
      character(len=IDD_MAX_TYPE_NAME)       :: profile_type
      character(len=IDD_MAX_DESCRIPTION)     :: description
      character(len=IDD_MAX_NAME)            :: algorithm
      character(len=IDD_MAX_VERSION)         :: algorithm_version
      character(len=IDD_MAX_DESCRIPTION)     :: algorithm_options
      !== tabulated form
      integer                                :: num_tabulated
      real(8), dimension(:), pointer         :: altitudes
      real(8), dimension(:), pointer         :: plasma_frequencies
      real(8), dimension(:), pointer         :: plasma_densities
      ! other profile values can be added here
      ! tabulated uncertainty bounds for each Ne value
      logical                                :: plasma_density_uncertainty_available
      real(8), dimension(:), pointer         :: plasma_density_bounds
      real(8), dimension(:), pointer         :: plasma_density_lower_bounds
      real(8), dimension(:), pointer         :: plasma_density_upper_bounds
      logical                                :: plasma_frequency_uncertainty_available
      real(8), dimension(:), pointer         :: plasma_frequency_bounds
      real(8), dimension(:), pointer         :: plasma_frequency_lower_bounds
      real(8), dimension(:), pointer         :: plasma_frequency_upper_bounds
      ! other uncertainties can be added here
      !== chebyshev representation
      type(IDD_PROFILE_COEFS)                :: f2_chebyshev
      type(IDD_PROFILE_COEFS)                :: f1_chebyshev
      type(IDD_PROFILE_COEFS)                :: e_chebyshev
      !== POLAN representation - ToDo
      type(IDD_PROFILE_COEFS)                :: f2_polan
      type(IDD_PROFILE_COEFS)                :: f1_polan
      type(IDD_PROFILE_COEFS)                :: fe_polan
      !== quasi-parabolic representation
      integer                                :: num_qp_segments
      real(8)                                :: earth_radius
      type(IDD_QP_SEGMENT), dimension(:), allocatable :: qps
      !== valley model
      character(len=IDD_MAX_MODEL_NAME)      :: valley_model
      real(8)                                :: valley_width_km
      real(8)                                :: valley_depth_pfu
      real(8)                                :: valley_start_height_km
      real(8)                                :: valley_start_freq_MHz
      !== topside Chapman profile
      type(IDD_TOPSIDE_CHAPMAN_LAYER)        :: top_chap

   end type IDD_PROFILE 

   type IONOGRAM_DERIVED_DATA
      type(IDD_TIMESTAMP_UTC)                        :: time
      type(IDD_LOCATION)                             :: location
      type(IDD_SYSTEM)                               :: system
      type(IDD_SOLAR_TERRESTRIAL_DATA)               :: geospace
      type(IDD_CHAR), dimension(IDD_TOTAL_CHARS)     :: chars
      type(IDD_TRACE), dimension(IDD_MAX_TRACES)     :: traces
      type(IDD_PROFILE), dimension(IDD_MAX_PROFILES) :: profiles
   end type IONOGRAM_DERIVED_DATA

include 'SAOXML_TAGS.inc'

	type(IONOGRAM_DERIVED_DATA)				:: idd

contains

   !
   ! ======================   f u n c t i o n s     ==================================
   !

subroutine idd_reset()
   integer                     :: i, j

   ! all optional metadata have to be cleared
   ! so as to avoid propagation of the metadata from a previous record
   ! (no guarantee for optional elements/attributes to be present in the source file)

   idd%time%year = NO_VALUE_YEAR
   idd%location%uml_sid = 0
   idd%location%ursi_code = ''
   idd%location%station_name = ''
   idd%location%geo_latitude = NO_VALUE_CHAR
   idd%location%geo_longitude = NO_VALUE_CHAR
   idd%system%comments = ''
   idd%system%autoscaled = .FALSE.
   idd%system%manual_scaler = ''
   idd%system%auto_scaler = ''
   idd%system%auto_scaler_version = ''
   idd%system%source = ''
   idd%system%source_type = ''
   idd%system%artist_flags = 0
   idd%system%ionogram%total_restricted_freq_bands = 0
   deallocate( idd%system%ionogram%upper_restricted_freq_limits, STAT=j )
   deallocate( idd%system%ionogram%lower_restricted_freq_limits, STAT=j )
   idd%system%ionogram%restricted_freq_units = ''
   idd%system%ionogram%frequency_stepping_rule = IDD_UNKNOWN_STEPPING
   idd%system%ionogram%start_frequency_MHz = NO_VALUE_CHAR
   idd%system%ionogram%stop_frequency_MHz = NO_VALUE_CHAR
   idd%system%ionogram%frequency_step = NO_VALUE_CHAR
   idd%system%ionogram%frequency_step_units = ''
   idd%system%ionogram%range_stepping_rule = IDD_UNKNOWN_STEPPING
   idd%system%ionogram%start_range_km = NO_VALUE_CHAR
   idd%system%ionogram%stop_range_km = NO_VALUE_CHAR
   idd%system%ionogram%range_step = NO_VALUE_CHAR
   idd%system%ionogram%range_step_units = ''
   idd%system%contact%available = .FALSE.
   idd%system%contact%name = ''
   idd%system%contact%affiliation = ''
   idd%system%contact%address = ''
   idd%system%contact%email = ''
   idd%system%digisonde_preface = ''
   idd%geospace%sunspot_number = NO_VALUE_CHAR
   idd%geospace%sunspot_number_status = ''
   idd%geospace%dip_angle_deg = NO_VALUE_CHAR
   idd%geospace%gyro_frequency_MHz = NO_VALUE_CHAR
   idd%geospace%magfield_model_altitude = NO_VALUE_CHAR
   idd%geospace%magfield_model_name = ''

   ! clear list of chars
   do i = 1, IDD_TOTAL_CHARS
      idd%chars(i)%value = NO_VALUE_CHAR
      idd%chars(i)%bound = NO_VALUE_CHAR
      idd%chars(i)%lower_bound = NO_VALUE_CHAR
      idd%chars(i)%upper_bound = NO_VALUE_CHAR
      idd%chars(i)%boundary_type = ''
      idd%chars(i)%q = ''
      idd%chars(i)%d = ''
      idd%chars(i)%units = ''
      idd%chars(i)%description = ''
      idd%chars(i)%flag = ''
   enddo
   do i = 1, IDD_TOTAL_MODELED_CHARS
      idd%system%model_name(i) = ''
      idd%system%model_options(i) = ''
   enddo

   ! clear list of traces
   do i = 1, IDD_MAX_TRACES
      idd%traces(i)%available = .FALSE.
      idd%traces(i)%amplitudes_available = .FALSE.
      idd%traces(i)%doppler_shifts_available = .FALSE.
      idd%traces(i)%num = 0   !  added 10/28/2010 by MAH
      deallocate( idd%traces(i)%frequencies, STAT=j )
      deallocate( idd%traces(i)%ranges, STAT=j )
      deallocate( idd%traces(i)%amplitudes, STAT=j )
      deallocate( idd%traces(i)%doppler_shifts, STAT=j )
      idd%traces(i)%trace_type = ''
      idd%traces(i)%layer = ''
      idd%traces(i)%polarization = ''
      idd%traces(i)%multiple = ''
      idd%traces(i)%amplitude_no_value = NO_VALUE_CHAR
      idd%traces(i)%doppler_shift_no_value = NO_VALUE_CHAR
   enddo

   ! clear list of profiles
   do i = 1, IDD_MAX_PROFILES
      idd%profiles(i)%available = .FALSE.
      idd%profiles(i)%num_tabulated = 0   !  added 10/28/2010 by MAH
      deallocate( idd%profiles(i)%altitudes, STAT=j )
      deallocate( idd%profiles(i)%plasma_frequencies, STAT=j )
      deallocate( idd%profiles(i)%plasma_densities, STAT=j )
      idd%profiles(i)%num_qp_segments = 0   !  added 10/28/2010 by MAH
      deallocate( idd%profiles(i)%qps, STAT=j )
      idd%profiles(i)%f2_chebyshev%num = 0
      deallocate( idd%profiles(i)%f2_chebyshev%c, STAT=j )
      idd%profiles(i)%f1_chebyshev%num = 0
      deallocate( idd%profiles(i)%f1_chebyshev%c, STAT=j )
      idd%profiles(i)%e_chebyshev%num = 0
      deallocate( idd%profiles(i)%e_chebyshev%c, STAT=j )
      ! clear valley model
      idd%profiles(i)%valley_model = ''
      idd%profiles(i)%valley_width_km = NO_VALUE_CHAR
      idd%profiles(i)%valley_depth_pfu = NO_VALUE_CHAR
      idd%profiles(i)%valley_start_height_km = NO_VALUE_CHAR
      idd%profiles(i)%valley_start_freq_MHz = NO_VALUE_CHAR
      ! reset topside model
      idd%profiles(i)%top_chap%available = .FALSE.
      idd%profiles(i)%top_chap%model = ''
      idd%profiles(i)%description = ''
      idd%profiles(i)%algorithm = ''
      idd%profiles(i)%algorithm_version = ''
      idd%profiles(i)%algorithm_options = ''
      ! reset uncertainty data
      idd%profiles(i)%plasma_density_uncertainty_available = .FALSE.
      deallocate( idd%profiles(i)%plasma_density_bounds, STAT=j )
      deallocate( idd%profiles(i)%plasma_density_lower_bounds, STAT=j )
      deallocate( idd%profiles(i)%plasma_density_upper_bounds, STAT=j )
      idd%profiles(i)%plasma_frequency_uncertainty_available = .FALSE.
      deallocate( idd%profiles(i)%plasma_frequency_bounds, STAT=j )
      deallocate( idd%profiles(i)%plasma_frequency_lower_bounds, STAT=j )
      deallocate( idd%profiles(i)%plasma_frequency_upper_bounds, STAT=j )

   enddo

end subroutine idd_reset


integer function idd_ursi_code_to_char_index( ursi_code )
   character*2, intent(in) :: ursi_code

   do idd_ursi_code_to_char_index = 1, IDD_TOTAL_STANDARD_CHARS
      if( URSI_CODE_CHAR(idd_ursi_code_to_char_index) .eq. ursi_code ) return 
   enddo
   idd_ursi_code_to_char_index = NO_VALUE_INDEX
end function idd_ursi_code_to_char_index

integer function idd_modeled_char_name_to_char_index( char_name )
   character(len=*), intent(in) :: char_name
   integer index
   do index = 1, IDD_TOTAL_MODELED_CHARS
      if( trim(MODELED_CHAR_NAME(index)) .eq. trim(char_name) ) then 
         idd_modeled_char_name_to_char_index = index + IDD_TOTAL_STANDARD_CHARS
         return
     endif
   enddo
   idd_modeled_char_name_to_char_index = NO_VALUE_INDEX
end function idd_modeled_char_name_to_char_index

integer function idd_custom_char_name_to_char_index( char_name )
   character(len=*), intent(in) :: char_name
   integer index
   do index = 1, IDD_TOTAL_CUSTOM_CHARS
      if( trim(CUSTOM_CHAR_NAME(index)) .eq. trim(char_name) ) then
         idd_custom_char_name_to_char_index = index + IDD_TOTAL_STANDARD_CHARS + IDD_TOTAL_MODELED_CHARS
         return
     endif
   enddo
   idd_custom_char_name_to_char_index = NO_VALUE_INDEX
end function idd_custom_char_name_to_char_index


real(8) function idd_get_char( index )
   integer, intent(in)                     :: index
   if( index .eq. NO_VALUE_INDEX .or. index .le. 0 .or. index > IDD_TOTAL_CHARS ) then
     idd_get_char = NO_VALUE_CHAR
   else
     idd_get_char = idd%chars(index)%value
   endif
end function idd_get_char

subroutine idd_set_char( index, value )
   integer, intent(in)                      :: index
   real(8), intent(in)                      :: value
   if( index .le. IDD_TOTAL_STANDARD_CHARS ) then
     call idd_set_char_with_units( index, value, trim(URSI_UNITS_CHAR(index)) )
   else
     call idd_set_char_with_units( index, value, NO_VALUE_UNITS )
   endif
end subroutine idd_set_char

subroutine idd_set_char_with_units( index, value, units )
   integer, intent(in)                      :: index
   real(8), intent(in)                      :: value
   character(len=*), intent(in)             :: units
   if( index .eq. NO_VALUE_INDEX .or. index .le. 0 .or. index > IDD_TOTAL_CHARS ) return
   idd%chars(index)%value = value
   idd%chars(index)%units = units
end subroutine idd_set_char_with_units

subroutine idd_set_char_bounds( index, bound, lower_bound, upper_bound, boundary_type )
   integer, intent(in )                     :: index
   real(8), intent(in)                      :: bound, lower_bound, upper_bound
   character(len=*), intent(in)             :: boundary_type
   real(8)                                  :: d1, d2
   if( index .ne. NO_VALUE_INDEX .and. index .ge. 0 .and. index .le. IDD_TOTAL_CHARS ) then
      ! set all values first 
      idd%chars(index)%bound = bound
      idd%chars(index)%lower_bound = lower_bound
      idd%chars(index)%upper_bound = upper_bound
      ! figure out which bounds are still missing
      ! possibly only symmetric bound is given
      if( lower_bound .eq. NO_VALUE_CHAR .and. bound .ne. NO_VALUE_CHAR) &
         idd%chars(index)%lower_bound = idd%chars(index)%value - bound
      if( upper_bound .eq. NO_VALUE_CHAR .and. bound .ne. NO_VALUE_CHAR) &
         idd%chars(index)%upper_bound = idd%chars(index)%value + bound
      ! bound may not be given, only upper-lower
      if( bound .eq. NO_VALUE_CHAR ) then
         if( lower_bound .eq. NO_VALUE_CHAR .and. upper_bound .ne. NO_VALUE_CHAR ) then
            idd%chars(index)%bound = idd%chars(index)%upper_bound - idd%chars(index)%value
         elseif( upper_bound .eq. NO_VALUE_CHAR .and. lower_bound .ne. NO_VALUE_CHAR ) then
            idd%chars(index)%bound = idd%chars(index)%value - idd%chars(index)%lower_bound
         elseif( upper_bound .ne. NO_VALUE_CHAR .and. lower_bound .ne. NO_VALUE_CHAR ) then
            d1 = idd%chars(index)%upper_bound - idd%chars(index)%value
            d2 = idd%chars(index)%value - idd%chars(index)%lower_bound
            idd%chars(index)%bound = max( d1, d2 )
         endif
      endif
      ! boundary type (optional)
      idd%chars(index)%boundary_type = boundary_type 
   endif
end subroutine idd_set_char_bounds

subroutine idd_set_char_qd_letters( index, q, d)
   integer, intent(in)                      :: index
   character, intent(in)                    :: q, d
   idd%chars(index)%q = q
   idd%chars(index)%d = d
end subroutine idd_set_char_qd_letters

subroutine idd_set_char_model( index, name, optionz)
   integer, intent(in)                      :: index
   character(len=*), intent(in)             :: name, optionz
   integer                                  :: local_index
   local_index = index - IDD_TOTAL_STANDARD_CHARS
   if( local_index > 0 .and. local_index < IDD_TOTAL_MODELED_CHARS + 1 ) then
      if( len_trim(name) > 0 ) idd%system%model_name(local_index) = trim(name)
      if( len_trim(optionz) > 0 ) idd%system%model_options(local_index) = trim(optionz)
   endif
end subroutine idd_set_char_model

!
!-----   T R A C E    s u p p o r t    f u n c t i o n s
!
type(IDD_TRACE) function idd_get_F2o_trace()
   idd_get_F2o_trace = idd_get_trace( VAL_LAYER_F2, 1, VAL_ORDINARY )
end function idd_get_F2o_trace

type(IDD_TRACE) function idd_get_F1o_trace()
   idd_get_F1o_trace = idd_get_trace( VAL_LAYER_F1, 1, VAL_ORDINARY )
end function idd_get_F1o_trace

type(IDD_TRACE) function idd_get_Eo_trace()
   idd_get_Eo_trace = idd_get_trace( VAL_LAYER_E, 1, VAL_ORDINARY )
end function idd_get_Eo_trace

type(IDD_TRACE) function idd_get_ESo_trace()
   idd_get_ESo_trace = idd_get_trace( VAL_LAYER_ES, 1, VAL_ORDINARY )
end function idd_get_ESo_trace

type(IDD_TRACE) function idd_get_trace( layer, multiple, polarization )
   character(len=*) :: layer
   integer          :: multiple
   character        :: polarization
   integer          :: i
   idd_get_trace = idd%traces(IDD_MAX_TRACES) ! this one is presumably not available... TBR
   do i = 1, IDD_MAX_TRACES
      if( idd%traces(i)%num .eq. 0 ) exit
     if( idd%traces(i)%layer .eq. layer .and. &
         idd%traces(i)%multiple .eq. multiple .and. &
         idd%traces(i)%polarization .eq. polarization ) then
        idd_get_trace = idd%traces(i)
        exit
     endif
   enddo
end function idd_get_trace

integer function idd_get_total_traces()
   integer                     :: i
   idd_get_total_traces = 0
   do i = 1, IDD_MAX_TRACES
      if( .not. idd%traces(i)%available ) exit
     idd_get_total_traces = idd_get_total_traces + 1
   enddo
end function idd_get_total_traces

integer function idd_get_total_profiles()
   integer                     :: i
   idd_get_total_profiles = 0
   do i = 1, IDD_MAX_PROFILES
      if( .not. idd%profiles(i)%available ) exit
     idd_get_total_profiles = idd_get_total_profiles + 1
   enddo
end function idd_get_total_profiles

real(8) function idd_frequency_to_density_cm3( frequency_MHz )
   real(8) :: frequency_MHz
   idd_frequency_to_density_cm3 = 1.E6 * frequency_MHz * frequency_MHz /80.6
end function idd_frequency_to_density_cm3

real(8) function idd_frequency_to_density_m3( frequency_MHz )
   real(8) :: frequency_MHz
   idd_frequency_to_density_m3 = 1.E12 * frequency_MHz * frequency_MHz /80.6
end function idd_frequency_to_density_m3

real(8) function idd_density_to_frequency( density_cm3 )
   real(8) :: density_cm3
   idd_density_to_frequency = sqrt( density_cm3 * 80.6 / 1.E6 )
end function idd_density_to_frequency

end module IonogramDerivedData

